@@ -14,8 +14,11 @@ from .history import History, InMemoryHistory
 from .search_state import SearchState
 from .selection import SelectionType, SelectionState
 from .utils import Callback
+from .cache import FastDictCache
 from .validation import ValidationError
 
+from six.moves import range
+
 import os
 import six
 import subprocess
@@ -29,6 +32,7 @@ __all__ = (
     'unindent',
 )
 
+
 class EditReadOnlyBuffer(Exception):
     " Attempt editing of read-only :class:`.Buffer`. "
 
@@ -42,8 +46,6 @@ class AcceptAction(object):
         :class:`~prompt_toolkit.interface.CommandLineInterface` and
         :class:`~prompt_toolkit.document.Document`. It is called when the user
         accepts input.
-    :param render_cli_done: When using a handler, first render the CLI in the
-        'done' state, then call the handler. This
     """
     def __init__(self, handler=None):
         assert handler is None or callable(handler)
@@ -225,6 +227,9 @@ class Buffer(object):
         self.on_text_insert = on_text_insert or Callback()
         self.on_cursor_position_changed = on_cursor_position_changed or Callback()
 
+        # Document cache. (Avoid creating new Document instances.)
+        self._document_cache = FastDictCache(Document, size=10)
+
         self.reset(initial_document=initial_document)
 
     def reset(self, initial_document=None, append_to_history=False):
@@ -273,10 +278,23 @@ class Buffer(object):
 
     def _set_text(self, value):
         """ set text at current working_index. Return whether it changed. """
-        original_value = self._working_lines[self.working_index]
-        self._working_lines[self.working_index] = value
+        working_index = self.working_index
+        working_lines = self._working_lines
 
-        return value != original_value
+        original_value = working_lines[working_index]
+        working_lines[working_index] = value
+
+        # Return True when this text has been changed.
+        if len(value) != len(original_value):
+            # For Python 2, it seems that when two strings have a different
+            # length and one is a prefix of the other, Python still scans
+            # character by character to see whether the strings are different.
+            # (Some benchmarking showed significant differences for big
+            # documents. >100,000 of lines.)
+            return True
+        elif value != original_value:
+            return True
+        return False
 
     def _set_cursor_position(self, value):
         """ Set cursor position. Return whether it changed. """
@@ -363,9 +381,10 @@ class Buffer(object):
     def document(self):
         """
         Return :class:`~prompt_toolkit.document.Document` instance from the
-        current text and cursor position.
+        current text, cursor position and selection state.
         """
-        return Document(self.text, self.cursor_position, selection=self.selection_state)
+        return self._document_cache[
+            self.text, self.cursor_position, self.selection_state]
 
     @document.setter
     def document(self, value):
@@ -555,7 +574,7 @@ class Buffer(object):
         assert self.selection_state
 
         # Get lines.
-        from_, to = self.document.selection_range()
+        from_, to = sorted([self.cursor_position, self.selection_state.original_cursor_position])
 
         before = self.text[:from_]
         lines = self.text[from_:to].splitlines()
@@ -800,23 +819,12 @@ class Buffer(object):
         """
         Copy selected text and return :class:`.ClipboardData` instance.
         """
-        if self.selection_state:
-            type = self.selection_state.type
+        new_document, clipboard_data = self.document.cut_selection()
+        if _cut:
+            self.document = new_document
 
-            # Take start and end of selection
-            from_, to = self.document.selection_range()
-
-            copied_text = self.text[from_:to]
-
-            # If cutting, remove the text and set the new cursor position.
-            if _cut:
-                self.document = Document(text=self.text[:from_] + self.text[to + 1:],
-                                         cursor_position=min(from_, to))
-
-            self.selection_state = None
-            return ClipboardData(copied_text, type)
-        else:
-            return ClipboardData('')
+        self.selection_state = None
+        return clipboard_data
 
     def cut_selection(self):
         """
@@ -824,6 +832,13 @@ class Buffer(object):
         """
         return self.copy_selection(_cut=True)
 
+    def paste_clipboard_data(self, data, before=False, count=1):
+        """
+        Insert the data from the clipboard.
+        """
+        assert isinstance(data, ClipboardData)
+        self.document = self.document.paste_clipboard_data(data, before=before, count=count)
+
     def newline(self, copy_margin=True):
         """
         Insert a line ending at the current position.
@@ -865,16 +880,21 @@ class Buffer(object):
         :param fire_event: Fire `on_text_insert` event. This is mainly used to
             trigger autocompletion while typing.
         """
+        # Original text & cursor position.
+        otext = self.text
+        ocpos = self.cursor_position
+
         # In insert/text mode.
         if overwrite:
-            # Don't overwrite the newline itself. Just before the line ending, it should act like insert mode.
-            overwritten_text = self.text[self.cursor_position:self.cursor_position+len(data)]
+            # Don't overwrite the newline itself. Just before the line ending,
+            # it should act like insert mode.
+            overwritten_text = otext[ocpos:ocpos + len(data)]
             if '\n' in overwritten_text:
                 overwritten_text = overwritten_text[:overwritten_text.find('\n')]
 
-            self.text = self.text[:self.cursor_position] + data + self.text[self.cursor_position+len(overwritten_text):]
+            self.text = otext[:ocpos] + data + otext[ocpos + len(overwritten_text):]
         else:
-            self.text = self.text[:self.cursor_position] + data + self.text[self.cursor_position:]
+            self.text = otext[:ocpos] + data + otext[ocpos:]
 
         if move_cursor:
             self.cursor_position += len(data)
@@ -883,31 +903,6 @@ class Buffer(object):
         if fire_event:
             self.on_text_insert.fire()
 
-    def paste_clipboard_data(self, data, before=False, count=1):
-        """
-        Insert the data from the clipboard.
-        """
-        assert isinstance(data, ClipboardData)
-
-        if data.type == SelectionType.CHARACTERS:
-            if before:
-                self.insert_text(data.text * count)
-            else:
-                self.cursor_right()
-                self.insert_text(data.text * count, fire_event=False)
-                self.cursor_left()
-
-        elif data.type == SelectionType.LINES:
-            if before:
-                self.cursor_position += self.document.get_start_of_line_position(after_whitespace=False)
-                self.insert_text((data.text + '\n') * count, move_cursor=False)
-            else:
-                self.cursor_position += self.document.get_end_of_line_position()
-                self.insert_text(('\n' + data.text) * count, move_cursor=False, fire_event=False)
-                self.cursor_down()
-
-            self.cursor_position += self.document.get_start_of_line_position(after_whitespace=True)
-
     def undo(self):
         # Pop from the undo-stack until we find a text that if different from
         # the current text. (The current logic of `save_to_undo_stack` will
@@ -1027,10 +1022,10 @@ class Buffer(object):
         # Do 'count' search iterations.
         working_index = self.working_index
         document = self.document
-        for i in range(count):
+        for _ in range(count):
             result = search_once(working_index, document)
             if result is None:
-                return # Nothing found.
+                return  # Nothing found.
             else:
                 working_index, document = result
 
@@ -1054,8 +1049,8 @@ class Buffer(object):
         Apply search. If something is found, set `working_index` and
         `cursor_position`.
         """
-        search_result = self._search(search_state,
-            include_current_position=include_current_position, count=count)
+        search_result = self._search(
+            search_state, include_current_position=include_current_position, count=count)
 
         if search_result is not None:
             working_index, cursor_position = search_result
