@@ -4,7 +4,7 @@ from prompt_toolkit.buffer import ClipboardData, indent, unindent
 from prompt_toolkit.document import Document
 from prompt_toolkit.enums import IncrementalSearchDirection, SEARCH_BUFFER, SYSTEM_BUFFER
 from prompt_toolkit.filters import Filter, Condition, HasArg, Always, to_cli_filter, IsReadOnly
-from prompt_toolkit.key_binding.vi_state import ViState, CharacterFind, InputMode
+from prompt_toolkit.key_binding.vi_state import CharacterFind, InputMode
 from prompt_toolkit.keys import Keys
 from prompt_toolkit.layout.utils import find_window_for_buffer_name
 from prompt_toolkit.selection import SelectionType
@@ -26,13 +26,21 @@ __all__ = (
 class ViStateFilter(Filter):
     """
     Filter to enable some key bindings only in a certain Vi input mode.
+
+    :param get_vi_state: Callable that takes a `CommandLineInterface` and
+        returns a :class:`~prompt_toolkit.key_binding.vi_state.ViState` instance.
     """
-    def __init__(self, vi_state, mode):
-        self.vi_state = vi_state
+    # Note: The reason for making get_vi_state a callable, is that this way,
+    #       the registry of key bindings becomes more stateless and can be
+    #       reused for multiple CommandLineInterface instances.
+    def __init__(self, get_vi_state, mode):
+        assert callable(get_vi_state)
+
+        self.get_vi_state = get_vi_state
         self.mode = mode
 
     def __call__(self, cli):
-        return self.vi_state.input_mode == self.mode
+        return self.get_vi_state(cli).input_mode == self.mode
 
 
 class CursorRegion(object):
@@ -54,15 +62,17 @@ class CursorRegion(object):
             return self.end, self.start
 
 
-def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None):
+def load_vi_bindings(registry, get_vi_state, enable_visual_key=Always(), get_search_state=None, filter=None):
     """
     Vi extensions.
 
     # Overview of Readline Vi commands:
     # http://www.catonmat.net/download/bash-vi-editing-mode-cheat-sheet.pdf
 
+    :param get_vi_state: Callable that takes a CommandLineInterface instances and returns the used ViState.
     :param enable_visual_key: Filter to enable lowercase 'v' bindings. A reason to disable these
          are to support open-in-editor functionality. These key bindings conflict.
+     :param get_search_state: None or a callable that takes a CommandLineInterface and returns a SearchState.
     """
     # Note: Some key bindings have the "~IsReadOnly()" filter added. This
     #       prevents the handler to be executed when the focus is on a
@@ -73,14 +83,18 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
     #       handled correctly. There is no need to add "~IsReadOnly" to all key
     #       bindings that do text manipulation.
 
-    assert isinstance(vi_state, ViState)
+    assert callable(get_vi_state)
     enable_visual_key = to_cli_filter(enable_visual_key)
 
+    # Default get_search_state.
+    if get_search_state is None:
+        def get_search_state(cli): return cli.search_state
+
     handle = create_handle_decorator(registry, filter)
 
-    insert_mode = ViStateFilter(vi_state, InputMode.INSERT) & ~ filters.HasSelection()
-    navigation_mode = ViStateFilter(vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
-    replace_mode = ViStateFilter(vi_state, InputMode.REPLACE) & ~ filters.HasSelection()
+    insert_mode = ViStateFilter(get_vi_state, InputMode.INSERT) & ~ filters.HasSelection()
+    navigation_mode = ViStateFilter(get_vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
+    replace_mode = ViStateFilter(get_vi_state, InputMode.REPLACE) & ~ filters.HasSelection()
     selection_mode = filters.HasSelection()
 
     vi_transform_functions = [
@@ -109,7 +123,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         if (
                 (filter is None or filter(event.cli)) and  # First make sure that this key bindings are active.
 
-                vi_state.input_mode == InputMode.NAVIGATION and
+                get_vi_state(event.cli).input_mode == InputMode.NAVIGATION and
                 buffer.document.is_cursor_at_the_end_of_line and
                 len(buffer.document.current_line) > 0):
             buffer.cursor_position -= 1
@@ -122,6 +136,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         Escape goes to vi navigation mode.
         """
         buffer = event.current_buffer
+        vi_state = get_vi_state(event.cli)
 
         if vi_state.input_mode in (InputMode.INSERT, InputMode.REPLACE):
             buffer.cursor_position += buffer.document.get_cursor_left_position()
@@ -232,19 +247,19 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
     @handle(Keys.Insert, filter=navigation_mode)
     def _(event):
         " Presing the Insert key. "
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('a', filter=navigation_mode & ~IsReadOnly())
             # ~IsReadOnly, because we want to stay in navigation mode for
             # read-only buffers.
     def _(event):
         event.current_buffer.cursor_position += event.current_buffer.document.get_cursor_right_position()
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('A', filter=navigation_mode & ~IsReadOnly())
     def _(event):
         event.current_buffer.cursor_position += event.current_buffer.document.get_end_of_line_position()
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('C', filter=navigation_mode & ~IsReadOnly())
     def _(event):
@@ -256,7 +271,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
 
         deleted = buffer.delete(count=buffer.document.get_end_of_line_position())
         event.cli.clipboard.set_text(deleted)
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('c', 'c', filter=navigation_mode & ~IsReadOnly())
     @handle('S', filter=navigation_mode & ~IsReadOnly())
@@ -273,7 +288,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         # But we delete after the whitespace
         buffer.cursor_position += buffer.document.get_start_of_line_position(after_whitespace=True)
         buffer.delete(count=buffer.document.get_end_of_line_position())
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('D', filter=navigation_mode)
     def _(event):
@@ -310,11 +325,11 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
 
     @handle('i', filter=navigation_mode & ~IsReadOnly())
     def _(event):
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('I', filter=navigation_mode & ~IsReadOnly())
     def _(event):
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
         event.current_buffer.cursor_position += event.current_buffer.document.get_start_of_line_position(after_whitespace=True)
 
     @handle('J', filter=navigation_mode)
@@ -334,7 +349,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         Search next.
         """
         event.current_buffer.apply_search(
-            event.cli.search_state, include_current_position=False,
+            get_search_state(event.cli), include_current_position=False,
             count=event.arg)
 
     @handle('N', filter=navigation_mode)
@@ -343,7 +358,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         Search previous.
         """
         event.current_buffer.apply_search(
-            ~event.cli.search_state, include_current_position=False,
+            ~get_search_state(event.cli), include_current_position=False,
             count=event.arg)
 
     @handle('p', filter=navigation_mode)
@@ -378,7 +393,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         Go to 'replace'-mode.
         """
-        vi_state.input_mode = InputMode.REPLACE
+        get_vi_state(event.cli).input_mode = InputMode.REPLACE
 
     @handle('s', filter=navigation_mode & ~IsReadOnly())
     def _(event):
@@ -388,7 +403,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         text = event.current_buffer.delete(count=event.arg)
         event.cli.clipboard.set_text(text)
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('u', filter=navigation_mode, save_before=(lambda e: False))
     def _(event):
@@ -402,6 +417,56 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         event.current_buffer.start_selection(selection_type=SelectionType.LINES)
 
+    @handle(Keys.ControlV, filter=navigation_mode)
+    def _(event):
+        " Enter block selection mode. "
+        event.current_buffer.start_selection(selection_type=SelectionType.BLOCK)
+
+    @handle('V', filter=selection_mode)
+    def _(event):
+        """
+        Exit line selection mode, or go from non line selection mode to line
+        selection mode.
+        """
+        selection_state = event.current_buffer.selection_state
+
+        if selection_state.type != SelectionType.LINES:
+            selection_state.type = SelectionType.LINES
+        else:
+            event.current_buffer.exit_selection()
+
+    @handle('v', filter=navigation_mode & enable_visual_key)
+    def _(event):
+        " Enter character selection mode. "
+        event.current_buffer.start_selection(selection_type=SelectionType.CHARACTERS)
+
+    @handle('v', filter=selection_mode)
+    def _(event):
+        """
+        Exit character selection mode, or go from non-character-selection mode
+        to character selection mode.
+        """
+        selection_state = event.current_buffer.selection_state
+
+        if selection_state.type != SelectionType.CHARACTERS:
+            selection_state.type = SelectionType.CHARACTERS
+        else:
+            event.current_buffer.exit_selection()
+
+    @handle(Keys.ControlV, filter=selection_mode)
+    def _(event):
+        """
+        Exit block selection mode, or go from non block selection mode to block
+        selection mode.
+        """
+        selection_state = event.current_buffer.selection_state
+
+        if selection_state.type != SelectionType.BLOCK:
+            selection_state.type = SelectionType.BLOCK
+        else:
+            event.current_buffer.exit_selection()
+
+
     @handle('a', 'w', filter=selection_mode)
     @handle('a', 'W', filter=selection_mode)
     def _(event):
@@ -437,7 +502,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         clipboard_data = event.current_buffer.cut_selection()
         event.cli.clipboard.set_data(clipboard_data)
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('y', filter=selection_mode)
     def _(event):
@@ -509,7 +574,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
             from_, _ = buffer.document.translate_index_to_position(from_)
             to, _ = buffer.document.translate_index_to_position(to)
 
-            indent(buffer, from_ - 1, to, count=event.arg)  # XXX: why does translate_index_to_position return 1-based indexing???
+            indent(buffer, from_, to + 1, count=event.arg)
 
     @handle('<', filter=selection_mode)
     def _(event):
@@ -524,7 +589,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
             from_, _ = buffer.document.translate_index_to_position(from_)
             to, _ = buffer.document.translate_index_to_position(to)
 
-            unindent(buffer, from_ - 1, to, count=event.arg)
+            unindent(buffer, from_, to + 1, count=event.arg)
 
     @handle('O', filter=navigation_mode & ~IsReadOnly())
     def _(event):
@@ -533,7 +598,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         event.current_buffer.insert_line_above(
                 copy_margin=not event.cli.in_paste_mode)
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('o', filter=navigation_mode & ~IsReadOnly())
     def _(event):
@@ -542,7 +607,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         event.current_buffer.insert_line_below(
                 copy_margin=not event.cli.in_paste_mode)
-        vi_state.input_mode = InputMode.INSERT
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle('~', filter=navigation_mode)
     def _(event):
@@ -563,7 +628,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         b = event.cli.current_buffer
 
-        search_state = event.cli.search_state
+        search_state = get_search_state(event.cli)
         search_state.text = b.document.get_word_under_cursor()
         search_state.direction = IncrementalSearchDirection.BACKWARD
 
@@ -577,7 +642,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         b = event.cli.current_buffer
 
-        search_state = event.cli.search_state
+        search_state = get_search_state(event.cli)
         search_state.text = b.document.get_word_under_cursor()
         search_state.direction = IncrementalSearchDirection.FORWARD
 
@@ -649,19 +714,6 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
                 if substring:
                     event.cli.clipboard.set_text(substring)
 
-            @handle('v', *keys, filter=navigation_mode & enable_visual_key)
-            def visual_handler(event):
-                """ Create visual handler. (Enter character selection mode.) """
-                region = func(event)
-                buffer = event.current_buffer
-
-                start, end = region.sorted()
-                end += buffer.cursor_position - 1
-
-                buffer.cursor_position += start
-                buffer.start_selection(selection_type=SelectionType.CHARACTERS)
-                buffer.cursor_position = end
-
             def create(delete_only):
                 """ Create delete and change handlers. """
                 @handle('cd'[delete_only], *keys, filter=navigation_mode & ~IsReadOnly())
@@ -686,7 +738,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
 
                     # Only go back to insert mode in case of 'change'.
                     if not delete_only:
-                        vi_state.input_mode = InputMode.INSERT
+                        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
             create(True)
             create(False)
@@ -837,7 +889,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         Go to next occurance of character. Typing 'fx' will move the
         cursor to the next occurance of character. 'x'.
         """
-        vi_state.last_character_find = CharacterFind(event.data, False)
+        get_vi_state(event.cli).last_character_find = CharacterFind(event.data, False)
         match = event.current_buffer.document.find(event.data, in_current_line=True, count=event.arg)
         return CursorRegion(match or 0)
 
@@ -847,7 +899,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         Go to previous occurance of character. Typing 'Fx' will move the
         cursor to the previous occurance of character. 'x'.
         """
-        vi_state.last_character_find = CharacterFind(event.data, True)
+        get_vi_state(event.cli).last_character_find = CharacterFind(event.data, True)
         return CursorRegion(event.current_buffer.document.find_backwards(event.data, in_current_line=True, count=event.arg) or 0)
 
     @change_delete_move_yank_handler('t', Keys.Any)
@@ -855,7 +907,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         Move right to the next occurance of c, then one char backward.
         """
-        vi_state.last_character_find = CharacterFind(event.data, False)
+        get_vi_state(event.cli).last_character_find = CharacterFind(event.data, False)
         match = event.current_buffer.document.find(event.data, in_current_line=True, count=event.arg)
         return CursorRegion(match - 1 if match else 0)
 
@@ -864,7 +916,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         Move left to the previous occurance of c, then one char forward.
         """
-        vi_state.last_character_find = CharacterFind(event.data, True)
+        get_vi_state(event.cli).last_character_find = CharacterFind(event.data, True)
         match = event.current_buffer.document.find_backwards(event.data, in_current_line=True, count=event.arg)
         return CursorRegion(match + 1 if match else 0)
 
@@ -876,6 +928,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         def _(event):
             # Repeat the last 'f'/'F'/'t'/'T' command.
             pos = 0
+            vi_state = get_vi_state(event.cli)
 
             if vi_state.last_character_find:
                 char = vi_state.last_character_find.character
@@ -921,7 +974,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         Moves to the start of the visible region. (Below the scroll offset.)
         Implements 'cH', 'dH', 'H'.
         """
-        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        w = find_window_for_buffer_name(event.cli, event.cli.current_buffer_name)
         b = event.current_buffer
 
         if w:
@@ -942,7 +995,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         Moves cursor to the vertical center of the visible region.
         Implements 'cM', 'dM', 'M'.
         """
-        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        w = find_window_for_buffer_name(event.cli, event.cli.current_buffer_name)
         b = event.current_buffer
 
         if w:
@@ -962,7 +1015,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         Moves to the end of the visible region. (Above the scroll offset.)
         """
-        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        w = find_window_for_buffer_name(event.cli, event.cli.current_buffer_name)
         b = event.current_buffer
 
         if w:
@@ -984,18 +1037,9 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         Scrolls the window to makes the current line the first line in the visible region.
         """
-        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        w = find_window_for_buffer_name(event.cli, event.cli.current_buffer_name)
         b = event.cli.current_buffer
-
-        if w and w.render_info:
-            # Calculate the offset that we need in order to position the row
-            # containing the cursor in the center.
-            cursor_position_row = b.document.cursor_position_row
-
-            render_row = w.render_info.input_line_to_screen_line.get(cursor_position_row)
-            if render_row is not None:
-                w.vertical_scroll = max(0, render_row)
-
+        w.vertical_scroll = b.document.cursor_position_row
 
     @handle('z', '-', filter=navigation_mode|selection_mode)
     @handle('z', 'b', filter=navigation_mode|selection_mode)
@@ -1003,34 +1047,39 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         Scrolls the window to makes the current line the last line in the visible region.
         """
-        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
-        b = event.cli.current_buffer
+        w = find_window_for_buffer_name(event.cli, event.cli.current_buffer_name)
 
-        if w and w.render_info:
-            # Calculate the offset that we need in order to position the row
-            # containing the cursor in the center.
-            cursor_position_row = b.document.cursor_position_row
-
-            render_row = w.render_info.input_line_to_screen_line.get(cursor_position_row)
-            if render_row is not None:
-                w.vertical_scroll = max(0, (render_row - w.render_info.window_height))
+        # We can safely set the scroll offset to zero; the Window will meke
+        # sure that it scrolls at least enough to make the cursor visible
+        # again.
+        w.vertical_scroll = 0
 
     @handle('z', 'z', filter=navigation_mode|selection_mode)
     def _(event):
         """
         Center Window vertically around cursor.
         """
-        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        w = find_window_for_buffer_name(event.cli, event.cli.current_buffer_name)
         b = event.cli.current_buffer
+        info = w.render_info
 
-        if w and w.render_info:
+        if w and info:
             # Calculate the offset that we need in order to position the row
             # containing the cursor in the center.
-            cursor_position_row = b.document.cursor_position_row
+            scroll_height = info.window_height // 2
 
-            render_row = w.render_info.input_line_to_screen_line.get(cursor_position_row)
-            if render_row is not None:
-                w.vertical_scroll = max(0, int(render_row - w.render_info.window_height / 2))
+            y = max(0, b.document.cursor_position_row - 1)
+            height = 0
+            while y > 0:
+                line_height = info.get_height_for_line(y)
+
+                if height + line_height < scroll_height:
+                    height += line_height
+                    y -= 1
+                else:
+                    break
+
+            w.vertical_scroll = y
 
     @change_delete_move_yank_handler('%')
     def _(event):
@@ -1052,7 +1101,7 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
 
         else:
             # Move to the corresponding opening/closing bracket (()'s, []'s and {}'s).
-            return CursorRegion(buffer.document.matching_bracket_position)
+            return CursorRegion(buffer.document.find_matching_bracket_position())
 
     @change_delete_move_yank_handler('|')
     def _(event):
@@ -1142,9 +1191,8 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         """
         @handle(*keys, filter=selection_mode)
         def _(event):
-            range = event.current_buffer.document.selection_range()
-            if range:
-                event.current_buffer.transform_region(range[0], range[1], transform_func)
+            for from_, to_ in event.current_buffer.document.selection_ranges():
+                event.current_buffer.transform_region(from_, to_, transform_func)
 
     for k, f in vi_transform_functions:
         create_selection_transform_handler(k, f)
@@ -1166,11 +1214,13 @@ def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None
         pass
 
 
-def load_vi_open_in_editor_bindings(registry, vi_state, filter=None):
+def load_vi_open_in_editor_bindings(registry, get_vi_state, filter=None):
     """
     Pressing 'v' in navigation mode will open the buffer in an external editor.
     """
-    navigation_mode = ViStateFilter(vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
+    assert callable(get_vi_state)
+
+    navigation_mode = ViStateFilter(get_vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
     handle = create_handle_decorator(registry, filter)
 
     @handle('v', filter=navigation_mode)
@@ -1178,11 +1228,11 @@ def load_vi_open_in_editor_bindings(registry, vi_state, filter=None):
         event.current_buffer.open_in_editor(event.cli)
 
 
-def load_vi_system_bindings(registry, vi_state, filter=None):
-    assert isinstance(vi_state, ViState)
+def load_vi_system_bindings(registry, get_vi_state, filter=None):
+    assert callable(get_vi_state)
 
     has_focus = filters.HasFocus(SYSTEM_BUFFER)
-    navigation_mode = ViStateFilter(vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
+    navigation_mode = ViStateFilter(get_vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
 
     handle = create_handle_decorator(registry, filter)
 
@@ -1191,8 +1241,8 @@ def load_vi_system_bindings(registry, vi_state, filter=None):
         """
         '!' opens the system prompt.
         """
-        event.cli.focus_stack.push(SYSTEM_BUFFER)
-        vi_state.input_mode = InputMode.INSERT
+        event.cli.push_focus(SYSTEM_BUFFER)
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle(Keys.Escape, filter=has_focus)
     @handle(Keys.ControlC, filter=has_focus)
@@ -1200,30 +1250,35 @@ def load_vi_system_bindings(registry, vi_state, filter=None):
         """
         Cancel system prompt.
         """
-        vi_state.input_mode = InputMode.NAVIGATION
+        get_vi_state(event.cli).input_mode = InputMode.NAVIGATION
         event.cli.buffers[SYSTEM_BUFFER].reset()
-        event.cli.focus_stack.pop()
+        event.cli.pop_focus()
 
     @handle(Keys.ControlJ, filter=has_focus)
     def _(event):
         """
         Run system command.
         """
-        vi_state.input_mode = InputMode.NAVIGATION
+        get_vi_state(event.cli).input_mode = InputMode.NAVIGATION
 
         system_buffer = event.cli.buffers[SYSTEM_BUFFER]
         event.cli.run_system_command(system_buffer.text)
         system_buffer.reset(append_to_history=True)
 
         # Focus previous buffer again.
-        event.cli.focus_stack.pop()
+        event.cli.pop_focus()
 
 
-def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name=SEARCH_BUFFER):
-    assert isinstance(vi_state, ViState)
+def load_vi_search_bindings(registry, get_vi_state, get_search_state=None,
+                            filter=None, search_buffer_name=SEARCH_BUFFER):
+    assert callable(get_vi_state)  # Callable that takes a CLI and returns a ViState.
+    assert get_search_state is None or callable(get_search_state)
+
+    if not get_search_state:
+        def get_search_state(cli): return cli.search_state
 
     has_focus = filters.HasFocus(search_buffer_name)
-    navigation_mode = ~has_focus & (ViStateFilter(vi_state, InputMode.NAVIGATION) | filters.HasSelection())
+    navigation_mode = ~has_focus & (ViStateFilter(get_vi_state, InputMode.NAVIGATION) | filters.HasSelection())
     handle = create_handle_decorator(registry, filter)
 
     @handle('/', filter=navigation_mode)
@@ -1233,11 +1288,11 @@ def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name=
         Vi-style forward search.
         """
         # Set the ViState.
-        event.cli.search_state.direction = IncrementalSearchDirection.FORWARD
-        vi_state.input_mode = InputMode.INSERT
+        get_search_state(event.cli).direction = IncrementalSearchDirection.FORWARD
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
         # Focus search buffer.
-        event.cli.focus_stack.push(search_buffer_name)
+        event.cli.push_focus(search_buffer_name)
 
     @handle('?', filter=navigation_mode)
     @handle(Keys.ControlR, filter=~has_focus)
@@ -1246,34 +1301,57 @@ def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name=
         Vi-style backward search.
         """
         # Set the ViState.
-        event.cli.search_state.direction = IncrementalSearchDirection.BACKWARD
+        get_search_state(event.cli).direction = IncrementalSearchDirection.BACKWARD
 
         # Focus search buffer.
-        event.cli.focus_stack.push(search_buffer_name)
-        vi_state.input_mode = InputMode.INSERT
+        event.cli.push_focus(search_buffer_name)
+        get_vi_state(event.cli).input_mode = InputMode.INSERT
 
     @handle(Keys.ControlJ, filter=has_focus)
     def _(event):
         """
         Apply the search. (At the / or ? prompt.)
         """
-        input_buffer = event.cli.buffers[event.cli.focus_stack.previous]
+        input_buffer = event.cli.buffers.previous(event.cli)
         search_buffer = event.cli.buffers[search_buffer_name]
 
         # Update search state.
         if search_buffer.text:
-            event.cli.search_state.text = search_buffer.text
+            get_search_state(event.cli).text = search_buffer.text
 
         # Apply search.
-        input_buffer.apply_search(event.cli.search_state)
+        input_buffer.apply_search(get_search_state(event.cli))
 
         # Add query to history of search line.
         search_buffer.append_to_history()
         search_buffer.reset()
 
         # Focus previous document again.
-        vi_state.input_mode = InputMode.NAVIGATION
-        event.cli.focus_stack.pop()
+        get_vi_state(event.cli).input_mode = InputMode.NAVIGATION
+        event.cli.pop_focus()
+
+    def incremental_search(cli, direction, count=1):
+        " Apply search, but keep search buffer focussed. "
+        # Update search_state.
+        search_state = get_search_state(cli)
+        direction_changed = search_state.direction != direction
+
+        search_state.text = cli.buffers[search_buffer_name].text
+        search_state.direction = direction
+
+        # Apply search to current buffer.
+        if not direction_changed:
+            input_buffer = cli.buffers.previous(cli)
+            input_buffer.apply_search(search_state,
+                                      include_current_position=False, count=count)
+
+    @handle(Keys.ControlR, filter=has_focus)
+    def _(event):
+        incremental_search(event.cli, IncrementalSearchDirection.BACKWARD, count=event.arg)
+
+    @handle(Keys.ControlS, filter=has_focus)
+    def _(event):
+        incremental_search(event.cli, IncrementalSearchDirection.FORWARD, count=event.arg)
 
     def search_buffer_is_empty(cli):
         """ Returns True when the search buffer is empty. """
@@ -1286,9 +1364,9 @@ def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name=
         """
         Cancel search.
         """
-        vi_state.input_mode = InputMode.NAVIGATION
+        get_vi_state(event.cli).input_mode = InputMode.NAVIGATION
 
-        event.cli.focus_stack.pop()
+        event.cli.pop_focus()
         event.cli.buffers[search_buffer_name].reset()
 
 
